package net.locusworks.common.configuration;

import static net.locusworks.common.Charsets.UTF_8;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Map;
import java.util.Properties;
import java.util.stream.Collectors;

import net.locusworks.common.immutables.Pair;
import net.locusworks.common.immutables.Unit;

/**
 * Properties manager class to help load and read properties
 * @author Isaac Parenteau
 * @version 2.0.0
 * @date 09/17/2023
 */
public class PropertiesManager {
  /**
   * Load a configuration from resource
   * @param clazz class loader
   * @param src source of the file
   * @return properties
   * @throws IOException Exception thrown the file can't be read
   */
  public static Properties loadConfiguration(Class<?> clazz, String src) throws IOException {
    InputStream is = clazz.getResourceAsStream(src);
    if (is == null) {
      is = clazz.getClassLoader().getResourceAsStream(src);
    }
    if (is == null) {
      return null;
    }
    BufferedReader br = new BufferedReader(new InputStreamReader(is, UTF_8));
    return loadConfiguration(br);
  }

  /**
   * Load configuration from a file. This method has been deprecated and may be removed in
   * future released. Please use java.nio.Path
   * @param file File to load
   * @return properties
   * @throws IOException Exception thrown the file can't be read
   */
  @Deprecated
  public static Properties loadConfiguration(File file) throws IOException {
    return loadConfiguration(file.toPath());
  }

  /**
   * Load configuration from a file.
   * @param path the path to the file
   * @return properties
   * @throws IOException Exception thrown the file can't be read
   */
  public static Properties loadConfiguration(Path path) throws IOException {
    if (Files.notExists(path)) {
      return new Properties();
    }

    try(BufferedReader br = Files.newBufferedReader(path)) {
      return loadConfiguration(br);
    }
  }

  /**
   * Load configuration from a buffered reader
   * @param reader Buffered reader to read the properties values from
   * @return properties
   * @throws IOException Exception thrown the file can't be read
   */
  public static Properties loadConfiguration(BufferedReader reader) throws IOException {
    Properties config = new Properties();
    config.load(reader);
    return config;
  }

  /**
   * Add configurations from one properties file to another
   * @param to Properties file to copy values to
   * @param from Properties file to copy values from
   * @return a map containing the results of the values added
   */
  public static Map<String, String> addConfiguration(Properties to, Properties from) {

    return from.entrySet()
        .stream()
        .filter(entry -> !to.containsKey(entry.getKey()))
        .map(entry -> {
          String key = entry.getKey().toString();
          String value = entry.getValue().toString();
          to.put(key, value);
          return new Pair<String, String>(key, value);
        })
        .collect(Collectors.toMap(Unit::getValue1, Pair::getValue2));
  }

  /**
   * Removes configuration values that are not present in the comparedTo
   * @param from Properties file to remove values from
   * @param comparedTo Properties file to compare to
   * @return a map containing the results of the values removed
   */
  public static Map<String, String> removeConfiguration(Properties from, Properties comparedTo) {

    return from.keySet()
        .stream()
        .filter(key -> !comparedTo.containsKey(key)) //only get the items that are not in the comparedTo properties
        .map(key -> new Pair<String, String>(String.valueOf(key), String.valueOf(from.get(key))))
        .toList() //Create a list of paired items (key value) of the items that were filtered
        .stream()
        .peek(pair -> { //remove those pairs from the from properties
          from.remove(pair.getValue1());
        })
        .collect(Collectors.toMap(Unit::getValue1, Pair::getValue2));
  }

  /**
   * <p>Save the properties file to disk.</p>
   * <p><b>This method has been depecreated and could be removed in future release.<br/>Please use java.nio.Path</b></p>
   * @param props Properties file to save
   * @param fileToSave File to save to
   * @param comment Any comments to add
   */
  @Deprecated
  public static void saveConfiguration(Properties props, File fileToSave, String comment) {
    saveConfiguration(props, fileToSave.toPath(), comment);

  }

  /**
   * <p>Save the properties file to disk.</p>
   * @param props Properties file to save
   * @param fileToSave File to save to
   * @param comment Any comments to add
   */
  public static void saveConfiguration(Properties props, Path fileToSave, String comment) {
    try(OutputStream fos = Files.newOutputStream(fileToSave)) {
      props.store(fos, comment == null ? "" : comment);
    } catch (IOException ex) {
      throw new RuntimeException(ex.getMessage(), ex);
    }
  }
}
