package net.locusworks.common.crypto;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigInteger;
import java.security.spec.EncodedKeySpec;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.RSAPublicKeySpec;
import java.util.Arrays;
import java.util.Base64;

import net.locusworks.common.Charsets;
import net.locusworks.common.io.IOUtils;

import static net.locusworks.common.utils.Checks.checkArguments;
import static net.locusworks.common.utils.Utils.get;
import static net.locusworks.common.utils.Utils.size;

public class SSHEncodedKeySpec extends EncodedKeySpec {
  
  private static final String SSH_MARKER = "ssh-rsa";

  public SSHEncodedKeySpec(byte[] encodedKey) {
    super(encodedKey);
  }
  
  public RSAPublicKeySpec convertToRSAPubKeySpec() throws InvalidKeySpecException {
    try {
      byte[] data = this.getEncoded();
      InputStream stream = new ByteArrayInputStream(data);
      Iterable<String> parts = Arrays.asList(IOUtils.toString(stream, Charsets.UTF_8).split(" "));
      
      checkArguments(size(parts) >= 2 && SSH_MARKER.equals(get(parts, 0)), "Bad format, should be: ssh-rsa AAB3...");
      stream = new ByteArrayInputStream(Base64.getDecoder().decode(String.valueOf(get(parts, 1))));
      String marker = new String(readLengthFirst(stream));
      checkArguments(SSH_MARKER.equals(marker), "Looking for marker %s but received %s", SSH_MARKER, marker);
      BigInteger publicExponent = new BigInteger(readLengthFirst(stream));
      BigInteger modulus = new BigInteger(readLengthFirst(stream));
      return new RSAPublicKeySpec(modulus, publicExponent);
    } catch (Exception ex) {
      throw new InvalidKeySpecException(ex);
    }
  }

  @Override
  public String getFormat() {
    return null;
  }
  
  private static byte[] readLengthFirst(InputStream in) throws IOException {
    int[] bytes = new int[] {in.read(), in.read(), in.read(), in.read()};
    int length = 0;
    int shift = 24;
    for (int i = 0; i < bytes.length; i++) {
      length += bytes[i] << shift;
      shift -= 8;
    }
    byte[] val = new byte[length];
    in.read(val);
    return val;
  }

}
