package net.locusworks.common.utils;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import static net.locusworks.common.utils.Checks.checkArguments;
import static net.locusworks.common.utils.Checks.checkNotNull;

/**
 * Class to help split a string in various ways.
 * By partition (fixed length split) or by sequence (look for a specific string sequence).
 * it can also split on new line or not.
 */
public class Splitter {
  private enum SplitterType {
    PARTITION,
    SEQUENCE
  }
  private String splitSeq;
  private boolean omitEmptyStrings = false;
  private int partition;
  private int limit;
  private final SplitterType splitterType;
  private static Splitter splitter;

  private Splitter(String seq) {
    this.splitSeq = seq;
    this.splitterType = SplitterType.SEQUENCE;
  }

  private Splitter(int partition) {
    this.partition = partition;
    this.splitterType = SplitterType.PARTITION;
  }

  /**
   * Remove empty string from the resulting lists
   * @return this
   */
  public Splitter omitEmptyStrings() {
    this.omitEmptyStrings = true;
    return this;
  }

  /**
   * Return a subset of the resulting list
   * @param limit how many items to retrieve
   * @return this
   */
  public Splitter withLimit(int limit) {
    this.limit = limit;
    return this;
  }

  /**
   * Return an array instead of a list
   * @param sentence the string sentence to split
   * @return this
   */
  public String[] splitToArray(String sentence) {
    List<String> list = split(sentence);
    return list.toArray(new String[0]);
  }

  /**
   * Split the string
   * @param sentence the string to split
   * @return the resulting list.
   */
  public List<String> split(String sentence) {
    checkArguments(!Utils.isEmptyString(sentence), "provided value is null or empty");
    List<String> list = new ArrayList<>();

    if (splitterType == SplitterType.PARTITION) {
      populateForFixedWidth(sentence, list);
    } else {
      populateForTrimmer(sentence, list);
    }

    return limit > 0 ? list.subList(0, limit) : list;
  }

  private void populateForFixedWidth(String sentence, List<String> list) {
    checkArguments(partition > 0, "Partition should be greater than 0");
    int strLength = sentence.length();
    for (int i = 0; i < strLength; i += partition) {
      list.add(sentence.substring(i, Math.min(strLength, i + partition)));
    }
  }

  private void populateForTrimmer(String sentence, List<String> list) {
    checkNotNull(splitSeq, "Split value provided was null");
    for (String s : sentence.split(splitSeq)) {
      if (omitEmptyStrings && s.trim().isEmpty())
        continue;
      list.add(s.trim());
    }
  }

  /**
   * Split the string on fixed length partitions
   * @param partition the length to split the string on
   * @return this
   */
  public static Splitter fixedLengthSplit(int partition) {
    splitter = new Splitter(partition);
    return splitter;
  }

  /**
   * Split the length on a specified string sequence
   * @param split the sequence to split
   * @return this
   */
  public static Splitter on(String split) {
    splitter = new Splitter(split);
    return splitter;
  }

  /**
   * Split on new line sequence
   * @return this
   */
  public static Splitter onNewLine() {
    return on("\\r?\\n");
  }

  /**
   * Split on spaces
   * @return this
   */
  public static Splitter onSpace() {
    return on(" ");
  }

  /**
   * Separator on what the key value is. return map
   * @param separator the separator value
   * @return this
   */
  public MapSplitter withKeyValueSeparator(String separator) {
    return new MapSplitter(this, separator);
  }

  public static class MapSplitter {

    private final Splitter splitter;
    private final String separator;
    private boolean skipInvalid = false;

    private MapSplitter(Splitter splitter, String separator) {
      checkNotNull(splitter, "Splitter cannot be null");
      checkArguments(!Utils.isEmptyString(separator),
        "Key value separator cannot be empty or null");
      this.splitter = splitter;
      this.separator = separator;
    }

    public MapSplitter skipInvalidKeyValues() {
      this.skipInvalid = true;
      return this;
    }

    public Map<String, String> split(String sentence) {
      checkArguments(!Utils.isEmptyString(sentence), "provided value is null or empty");
      Map<String, String> map = new LinkedHashMap<>();

      for (String s : splitter.split(sentence)) {
        String[] keyValue = s.split(separator);
        try {
          checkArguments(keyValue.length == 2, "invalid length found for key value mapping");
        } catch (IllegalArgumentException ex) {
          if (!skipInvalid)
            throw ex;
          continue;
        }
        map.put(keyValue[0], keyValue[1]);
      }

      return map;
    }
  }
}
