package net.locusworks.common.utils;

import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Modifier;
import java.util.*;
import java.util.Map.Entry;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Consumer;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import net.locusworks.common.annotations.MapValue;
import net.locusworks.common.interfaces.ThrowingConsumer;

/***
 * Utils class with generic methods
 * @author Isaac Parenteau
 * @version 1.0.0
 * @date 02/15/2018
 */
public class Utils {


  /**
   * Finds if all the values are equal
   * @param or first value is equal to at least one other value
   * @param values the values to compare
   * @return true if the first value is equal to at least one other value; false otherwise
   */
  @SafeVarargs
  public static <E> boolean areEqual(Boolean or, E... values) {
    if (values.length < 2) {
      throw new IllegalArgumentException("Not enough values to compare");
    }

    E firstVal = values[0];
    boolean equal = !or;
    
    for (int i = 1; i < values.length; i++) {
      if (or)
        equal |= firstVal.equals(values[i]);
      else
        equal &= firstVal.equals(values[i]);
    }

    return equal;
  }

  /**
   * Checks to see if values entered are all equal
   * @param <E>    the type parameter
   * @param values the values
   *
   * @return boolean
   */
  @SafeVarargs
  public static <E> boolean areEqual(E... values) {
    return areEqual(false, values);
  }

  /**
   * Checks to see if a group of values are not valid. All values have to be invalid to return true
   * One valid value will return false
   * @param values Values to check
   * @param <V> the expected class type of the objects passed in. All objects need to be of same type
   * @return true if all values are invalid, false otherwise
   */
  @SuppressWarnings("unchecked")
  public static <V> boolean areNotValid(V... values) {
    return !areValid(values);
  }

  /**
   * Checks to see if a group of values are valid. All values have to be valid to return true
   * One invalid value will return false
   * @param <V> the expected class type of the objects passed in. All objects need to be of same type
   * @param values the values to check
   * @return true if the values are valid, false otherwise
   */
  @SuppressWarnings("unchecked")
  public static <V> boolean areValid(V... values) {
    return validateValues(values);
  }

  /**
   * Build a map class
   *
   * @param <K>      the type parameter
   * @param <V>      the type parameter
   * @param mapClass The map class to map to i.e. HashMap, TreeMap etc
   * @param mapKey   the map key
   * @param mapValue the map value
   * @param data     The data to place in the map
   *
   * @return map The map with the objects
   */
  @SuppressWarnings("unchecked")
  public static <K, V> Map<K, V> buildMap(Class<?> mapClass, Class<K> mapKey, Class<V> mapValue,  Object... data) {
    Objects.requireNonNull(mapKey, "Null key value");
    Objects.requireNonNull(mapValue, "Null map value");

    Map<K, V> results;
    try {
      results = (Map<K, V>) mapClass.getDeclaredConstructor().newInstance();
    } catch (Exception ex) {
      throw new IllegalArgumentException("Unable to create instance of " + mapClass.getSimpleName() + " e: " + ex.getMessage());
    }

    if (data.length % 2 != 0) {
      throw new IllegalArgumentException("Odd number of arguments provided");
    }

    for(int i = 0; i < data.length; i+=2) {
      Object key = data[i];
      Object value = data[i + 1];

      if (!mapKey.isInstance(key)) {
        throw new IllegalArgumentException("Key is not the correct instance. Expecting " + mapKey.getName() + " Received "  + key.getClass().getName());
      }

      if (!mapValue.isInstance(value)) {
        throw new IllegalArgumentException("Value is not the correct instance. Expecting " + mapValue.getName() + " Received "  + value.getClass().getName());
      }
      results.put((K) key, (V) value);

    }
    return results;
  }

  /**
   * Builds a set of objects
   * @param setClass Class type of the set to create
   * @param setValue Class type of the objects being placed in the set
   * @param <E> Class type of the return object (should be the same as setValue)
   * @param data Data to place inside the set
   * @return Set filled with the data
   * @deprecated since JDk11 can use Set.of
   */
  @SuppressWarnings("unchecked")
  public static <E> Set<E> buildSet(Class<?> setClass, Class<E> setValue, E... data) {
    return buildSet(data);
  }

  /**
   * Builds a set of objects
   * @param data the data to set
   * @return the set
   * @param <E> the type
   */
  @SafeVarargs public static <E> Set<E> buildSet(E... data) {
    return Set.of(data);
  }

  /**
   * Public method of type HashMap called "buildStringHashMap".
   * The method takes String parameters and a data parameter.
   * The method returns a new HashMap that uses String types as key / value pairs.
   * @param data - The data
   * @return     - Returns a HashMap using the data.
   */
  public static HashMap<String, String> buildStringHashMap(String... data) {
    return new HashMap<>(buildMap(HashMap.class, String.class, String.class, (Object[]) data));
  }

  /**
   * Clone list and makes it unmodifiable.
   *
   * @return unmodifiable list
   */
  public static <E> List<E> cloneList(List<E> list) {
      return Collections.unmodifiableList(list);
  }

  /**
   * Clone object object.
   * @param obj the obj
   * @return the object
   * @throws IllegalAccessException thrown when the filed cannot be access
   * @throws InstantiationException Thrown when the object cannot be instantiated
   */
  @SuppressWarnings("unchecked")
  public static <O> O cloneObject(O obj)
    throws InstantiationException, IllegalAccessException, NoSuchMethodException, InvocationTargetException {
    Object clone = obj.getClass().getDeclaredConstructor().newInstance();
    for (Field field : obj.getClass().getDeclaredFields()) {
      field.setAccessible(true);
      if (field.get(obj) == null || Modifier.isFinal(field.getModifiers())){
        continue;
      }

      Class<?> type = field.getType();

      if (isPrimitiveOrWrapper(type)) {
        field.set(clone, field.get(obj));
      } else {
        field.set(clone, cloneObject(field.get(obj)));
      }
    }
    return (O) clone;
  }

  /***
   * Convert an object to a map. Checks to see if there is a MapValue annotation
   * and uses the values found to convert the fields to the key and the field value as the value
   * @param obj Object to convert
   * @return A map representation of the object
   * @throws Exception exception
   */
  public static Map<String, Object> convertToMap(Object obj) throws Exception {
    Class<?> clazz = obj.getClass();
    Map<String, Object> map = new LinkedHashMap<>();



    for (Field field : clazz.getDeclaredFields()) {
      field.setAccessible(true);
      boolean hasAnnotations = field.isAnnotationPresent(MapValue.class);

      String key = field.getName();
      Object value = field.get(obj);
      if (value == null) continue;

      if (hasAnnotations) {
        MapValue annotation = field.getDeclaredAnnotation(MapValue.class);
        if (annotation.ignore()) {
          continue;
        }

        if (!isEmptyString(annotation.value())) {
          key = annotation.value();
        }
      }
      map.put(key, value);
    }

    return map;
  }

  /**
   * Convert a generic map to Stringed map where the key and value are both strings
   * @param map map to convert 
   * @param <K> The expected class of the key
   * @param <V> The expected class of the value
   * @return convertedMap
   */
  public static <K, V> Map<String, String> convertToStringMap(Map<K, V> map) {
    Map<String, String> convert = new LinkedHashMap<>();
    for (Entry<K, V> item : map.entrySet()) {
      convert.put(String.valueOf(item.getKey()), String.valueOf(item.getValue()));
    }

    return convert;
  }

  /**
   * Converts an object toa  string map;
   * @param obj the object to convert
   * @return the map
   * @throws Exception thrown if something happens
   */
  public static Map<String, String> convertToStringMap(Object obj) throws Exception {
    return convertToStringMap(convertToMap(obj));
  }

  /***
   * Takes an input list and converts it to another list denoted by what is grabbed from the fieldName
   * @param fieldName Name of the field in the object to retrieve
   * @param list the list to retrieve values from
   * @param <E> The expected class of the List
   * @param <K> The expected class of the Iterable list
   * @return The converted list
   */
  public static <E, K> Set<E> extractFieldToSet(String fieldName, Iterable<K> list) {
    return listToSet(extractFieldToList(fieldName, list));
  }

  /**
   * Extract a field from an iterable object and place them in a set
   * @param fieldName field to check
   * @param list list to iterate
   * @param <E> The expected class of the Set
   * @param <K> The Expected class of the iterable list
   * @return set with the desired items
   */
  @SuppressWarnings("unchecked")
  public static <E, K> List<E> extractFieldToList(String fieldName, Iterable<K> list) {
    List<E> newSet = new ArrayList<>();
    Field field;
    for (K item : Utils.safeIterable(list)) {
      try {
        field = getField(item.getClass(), fieldName);

        if (field == null) {
          throw new IllegalArgumentException("Unable to find field with name " + fieldName);
        }

        E val = (E)field.get(item);
        newSet.add(val);

      } catch (Exception ex) {
        throw new IllegalArgumentException(ex);
      }
    }
    return newSet;
  }

  /**
   * Filters out items in a list with the given parameters
   * @param fieldName the field name to inspect for the filter value
   * @param filter the filter value to retrieve
   * @param list the list to inspect
   * @param <E> the expected class of the iterable list
   * @return List of items that met the criteria of the filter
   */
  @SuppressWarnings("unchecked")
  public static <E> List<E> filterList(String fieldName, Object filter, Iterable<E> list) {
    List<E> newList = new ArrayList<>();

    Field field;

    for (E item : safeIterable(list)) {
      try {
        field = getField(item.getClass(), fieldName);

        if (field == null) {
          throw new IllegalArgumentException("Unable to find field with name " + fieldName);
        }

        E val = (E)field.get(item);
        if (val.equals(filter)) {
          newList.add(item);
        }
      } catch (Exception ex) {
        throw new IllegalArgumentException(ex);
      }
    }

    return newList;
  }

  /**
   * Find a value within the list.
   * Public method of type E called "findValue".
   * The method takes a parameter of type E called "itemToFind" and a String parameter called "fieldName" and a parameter of type List called "list".
   * The method loops through each item in the list and tries to get the class name and return it as an item.
   * The method the takes each item and searches for it in the list.
   * @param itemToFind - The Item to find
   * @param fieldName  - The field name where the item is in
   * @param list       - The list to search
   * @param <E>        - The expected class of item to be found
   * @param <K>        - The expected class of the Iterable list
   * @return           - Returns the value of the specified field or null if there is no value.
   * 
   */
  @SuppressWarnings({"unlikely-arg-type" })
  public static <E, K> E findValue(E itemToFind, String fieldName, List<K> list) {
    return findValues(itemToFind, fieldName, list).stream().findFirst().orElse(null);
  }

  /**
   * Find values list.
   * Public method of type List called "findValues".
   * The method takes an E type parameter called "valueToFind" and a String parameter called "fieldName" and a List parameter called "list".
   * The method creates a new array list called "tmpList".
   * The method then iterates through the list and gets all the classes as items.
   * If it can't find the fields, it will throw an illegal argument exception.
   * @param valueToFind - The value to find
   * @param fieldName   - The field name
   * @param list        - The list
   * @param <E>         - The expected class of item to be found
   * @param <K>         - The expected class of the Iterable list
   * @return            - Returns the list.
   */
  @SuppressWarnings({"unlikely-arg-type" })
  public static <E, K> List<E> findValues(E valueToFind, String fieldName, List<K> list) {
    List<E> extracted = extractFieldToList(fieldName, list);
    return extracted.stream()
      .filter(Objects::nonNull)
      .filter(v -> v.equals(valueToFind)).collect(Collectors.toList());
  }

  /**
   * Use reflection to get the field values
   * @param clazz the class to find the field
   * @param fieldName the field name
   * @return field
   */
  private static Field getField(Class<?> clazz, String fieldName) {
    Field field = null;
    while(clazz != null) {
      try {
        field = clazz.getDeclaredField(fieldName);
        field.setAccessible(true);
        break;
      } catch(Exception ex) {
        clazz = clazz.getSuperclass();
      }
    }

    return field;
  }

  /**
   * A replacement for String.format. Allows for to many parameters or too few
   * Replaces {} in the string in order.
   *
   * @param message the message
   * @param args    the args
   *
   * @return string
   */
  public static String formatString(String message, Object... args) {
    String tmp = message;

    for(Object arg : args) {
      tmp = tmp.replaceFirst("\\{\\}", String.valueOf(arg));
    }

    return tmp;
  }

  /**
   * Get a class name
   *
   * @param <V> the type parameter
   * @param obj the obj
   *
   * @return class name
   */
  public static <V> String getClassName(V obj) {
    return getClassName(obj, false);
  }

  /**
   * Get a class name
   *
   * @param <V>     the type parameter
   * @param obj     the obj
   * @param verbose the verbose
   *
   * @return class name
   */
  public static <V> String getClassName(V obj, boolean verbose) {
    return obj == null ? "Unknown" : (verbose ? obj.getClass().getName() : obj.getClass().getSimpleName());
  }

  /**
   * Get a value from a map
   *
   * @param <T> the type parameter
   * @param <K> the type parameter
   * @param map the map with values
   * @param key the key to find
   *
   * @return map value
   */
  public static <T, K> T getMapValue(Map<K, T> map, K key) {
    return getMapValue(map, key, null);
  }

  /**
   * Get a value from a map
   *
   * @param <T>          the type parameter
   * @param <K>          the type parameter
   * @param map          the map with the values
   * @param key          the key to find
   * @param defaultValue the default value if the key is not found
   *
   * @return map value
   */
  public static <T, K> T getMapValue(Map<K, T> map, K key, T defaultValue) {
    return map.getOrDefault(key, defaultValue);
  }

  /**
   * Checks to see if the proved string is empty
   * new String(null)        = true
   * new String("")          = true
   * new String(" ")         = true
   * new String("foo")       = false
   * new String("   bar   ") = false
   * @param string String to check
   * @return true if the string is empty false otherwise
   */
  public static boolean isEmptyString(String string) {
    return string == null || string.trim().isEmpty();
  }

  /**
   * Checks to see if a passed in value is not a valid value
   * For Strings checks to see if they are empty or not null
   * For collections checks to see if they are null or empty
   * for all others just checks if they are null
   * @param <V> the expected class type of the object
   * @param values The value to check
   * @return true if the value is not valid, false otherwise
   */
  @SafeVarargs
  public static <V> boolean isNotValid(V... values) {
    return !validateValues(values);
  }
  
  /**
   * Checks to see if a passed in value is a valid value
   * For Strings checks to see if they are not empty or not null
   * For collections checks to see if they are not null and not empty
   * for all others just checks if they are not null
   * @param <V> the expected class type of the object
   * @param values The value to check
   * @return true if the value is valid, false otherwise
   */
  @SafeVarargs
  public static <V> boolean isValid(V... values) {
    return validateValues(values);
  }

  /**
   * Is primitive or wrapper boolean.
   *
   * @param clazz the clazz
   *
   * @return the boolean
   */
  public static boolean isPrimitiveOrWrapper(Class<?> clazz) {
    return clazz.isPrimitive() || getWrapperTypes().contains(clazz);
  }

  /**
   * Converts a list of items into a map.
   *
   * @param <K>          the type parameter
   * @param <V>          the type parameter
   * @param keyFieldName the name of the field that will be used as the key
   * @param list         the list to convert
   *
   * @return a map
   */
  @SuppressWarnings("unchecked")
  public static <K, V> Map<K, V> listToMap(String keyFieldName, List<V> list) {
    Map<K, V> map = new HashMap<>();

    for(V value : Utils.safeList(list)) {
      try {
        Field field = getField(value.getClass(), keyFieldName);
        K key = (K) field.get(value);
        map.put(key, value);
      } catch(Exception e) {
        throw new IllegalArgumentException(String.format("Unable to find field %s. -> %s", keyFieldName, e.getMessage()));
      }
    }

    return map;
  }

  /**
   * List to set.
   *
   * @param <E>      the type parameter
   * @param valueSet the value set
   *
   * @return the set
   */
  public static <E> Set<E> listToSet(Collection<E> valueSet) {
    return toSet(valueSet);
  }

  /**
   * Creates an array from a collection
   * Public method of type array called "makeArray".
   * The method takes a Collection object parameter called "collection" and a Class object parameter called "clazz".
   * The method first checks to see if the Collection object is null. If so, it returns null.
   * The method then creates an array called "results" and populates it with a new instance of "clazz" and the size of the collection.
   * The method then creates an Int variable called "index" and sets the value to zero.
   * The method then iterates through the collection and adds all the items to the results array.
   * The method then returns the results array.
   * @param <E>        - The type parameter
   * @param collection - The collection
   * @param clazz      - The clazz
   * @return e[]       - Returns the array of results.
   */
  @SuppressWarnings("unchecked")
  public static <E> E[] makeArray(Collection<E> collection, Class<E> clazz) {
    if (collection == null) {
      return null;
    }

    E[] results = (E[])Array.newInstance(clazz, collection.size());

    int index = 0;
    for (E item : collection) {
      results[index++] = item;
    }

    return results;
  }

  /**
   * Creates a list from a collection.
   * Public method of type List called "makeList".
   * The method takes a Collection object parameter called "collection".
   * The method checks to see if the Collection Object is null. If so, it returns null.
   * Otherwise, it creates an ArrayList object called "list".
   * It then iterates through the Collection Object and adds all the collection items to that list.
   * The method then returns the list.
   * @param <E>        - The type parameter
   * @param collection - The collection
   * @return           - Returns the list.
   */
  public static <E> List<E> makeList(Collection<E> collection) {
    if (collection == null) {
      return null;
    }
    return new ArrayList<>(collection);
  }

  /**
   * Public method of type List called "makeList".
   * The method takes a parameter of type E array called "array".
   * The method checks to see if the array is null. If so, it returns null.
   * Otherwise, it creates an ArrayList object called "newList".
   * It then iterates through the array and adds all the items to the new list.
   * The method then returns the list.
   * @param <E>   - The type parameter
   * @param array - The array
   * @return      - Returns the list.
   */
  public static <E> List<E> makeList(E[] array) {
    if (array == null) {
      return null;
    }

    return List.of(array);
  }

  /**
   * Creates a list from an iterable object
   * Public method of type List called "makeList".
   * The method takes an Iterable object parameter called "iter".
   * The method checks to see if the object is null. If so, it returns null.
   * Otherwise, it creates a new ArrayList object called "list".
   * It then iterates through the Iterable object and adds items to it.
   * After that it returns the list.
   * @param iter - The object to iterate over.
   * @param <E>  - The expected class of the Iterable object
   * @return     - Returns the list.
   */
  public static <E> List<E> makeList(Iterable<? extends E> iter) {
    if (iter == null) {
      return null;
    }
    List<E> list = new ArrayList<>();
    for (E item : iter) {
      list.add(item);
    }
    return list;
  }

  /**
   * Public method of type Set called "makeSet".
   * The method takes an Iterable object parameter called "collection".
   * The method first creates a new HashSet called "set".
   * The method then iterates through the collection and adds all the items from it to the HashSet.
   * The method then returns the HashSet.
   * @param <E>        - The type parameter
   * @param collection - The iterable object name.
   * @return           - Returns the HashSet with all the items of the collection inside of it.
   */
  public static <E> Set<E> makeSet(Iterable<E> collection) {
    Set<E> set = new HashSet<>();
    for (E item : safeIterable(collection)) {
      set.add(item);
    }
    return set;
  }

  /**
   * Map to list.
   *
   * @param <K> the type parameter
   * @param <V> the type parameter
   * @param map the map
   *
   * @return the list
   */
  public static <K, V> List<V> mapToList(Map<K, V> map) {

    return new ArrayList<>(map.values());
  }

  /**
   * Map to set.
   *
   * @param <K> the type parameter
   * @param <V> the type parameter
   * @param map the map
   *
   * @return the set
   */
  public static <K, V> Set<V> mapToSet(Map<K, V> map) {

    return new HashSet<>(map.values());
  }

  /**
   * Reverse a map switching key value pairs
   *
   * @param <K> the type parameter
   * @param <V> the type parameter
   * @param map the map
   *
   * @return map with reverse key value pairs
   */
  public static <K,V> Map<K, V> reverseMap(Map<V, K> map) {
    return map.entrySet().stream().collect(Collectors.toMap(Entry::getValue, Entry::getKey));
  }

  /**
   * create safe array e [ ].
   *
   * @param <E>  the type parameter
   * @param list the list
   *
   * @return the e [ ]
   */
  @SuppressWarnings("unchecked")
  public static <E> E[] safeArray(E[] list) {
    if (!validateValue(list)) {
      return (E[]) new Object[0];
    }
    return list;
  }


  /**
   * Creates a safe list
   *
   * @param <E>  the type parameter
   * @param list The list to check
   *
   * @return an empty list if list is invalid or the list if its valid
   */
  public static <E> Collection<E> safeList(Collection<E> list) {
    if (!validateValue(list)) {
      return new ArrayList<>();
    }
    return list;
  }


  /**
   * Creates a safe list
   *
   * @param <E>  the type parameter
   * @param list The list to check
   *
   * @return an empty list if list is invalid or the list if its valid
   */
  public static <E> Iterable<E> safeIterable(Iterable<E> list) {
    if (!validateValue(list)) {
      return new ArrayList<>();
    }
    return list;
  }

  /**
   * Creates a safe set
   * @param <E> the type parameter
   * @param set the set to check
   * @return an empty set if the set is null otherwise the set
   */
  public static <E> Set<E> safeSet(Set<E> set) {
    if (set == null) {
      return new HashSet<>();
    }
    return set;
  }

  /**
   * Checks to see if a string is not blank or null. 
   * if it's not blank it will return the string
   * otherwise it will return an empty string
   * @param string The string to check
   * @return the passed in string if it's not null either empty string
   */
  public static String safeString(String string) {
    return isEmptyString(string) ? "" : string;
  }

  /**
   * Sets to list.
   *
   * @param <E>      the type parameter
   * @param valueSet the value set
   *
   * @return the to list
   */
  public static <E> List<E> setToList(Set<E> valueSet) {
    return new ArrayList<>(valueSet);
  }

  /**
   * Converts a String into an integer without exception
   * @param value The string value to convert
   * @param defaultValue The default value to return if the string cant be converted
   * @return the integer representation of the passed in string or the default value if an exception occurred
   */
  public static Integer toInteger(String value, Integer defaultValue) {
    try {
      return Integer.parseInt(value);
    } catch (Exception ex) {
      return defaultValue;
    }
  }

  /**
   * Create a list from given values
   *
   * @param <E>    the type parameter
   * @param values the values to add the list
   *
   * @return List containing said values
   */
  @SafeVarargs
  public static <E> List<E> toList(E... values) {
    return Arrays.asList(values);
  }

  public static <E> List<E> toList(Iterable<E> iterable) {
    List<E> list = new ArrayList<>();
    for (E item : iterable) {
      list.add(item);
    }
    return list;
  }

  /**
   * Create a set from given values
   *
   * @param <E>    the type parameter
   * @param values the values to add the set
   *
   * @return Set containing said values
   */
  @SafeVarargs
  public static <E> Set<E> toSet(E... values) {
    return new LinkedHashSet<>(Arrays.asList(values));
  }
  
  public static List<Byte> toByteList(byte[] bytes) {
    return IntStream.range(0, bytes.length)
        .mapToObj(index -> bytes[index])
        .collect(Collectors.toList());
  }

  /**
   * Converts an iterable object to a hash set
   * @param iterable The Iterable object to convert
   * @param <E> the class type of the set 
   * @return set with the iterable items inside
   */
  public static <E> Set<E> toSet(Iterable<E> iterable) {
    Set<E> set = new HashSet<>();
    for (E value : iterable) {
      set.add(value);
    }
    return set;
  }
  
  public static <E> long size(Iterable<E> iterable) {
    long count = 0;
    for (Iterator<E> iter = iterable.iterator(); iter.hasNext(); iter.next()) { count++; }
    return count;
  }
  
  public static <E> E get(Iterable<E> iterable, int index) {
    AtomicInteger count = new AtomicInteger(0);
    Optional<E> test = StreamUtils.asStream(iterable)
        .filter(item -> count.getAndIncrement() == index)
        .findFirst();
    return test.orElse(null);
  }
  
  public static <T, E extends Exception> Consumer<T> handleExceptionWrapper(ThrowingConsumer<T, E> consumer) {
    return i -> {
      try {
        consumer.accept(i);
      } catch (Exception ex) {
        throw new RuntimeException(ex);
      }
    };
  }

  /**
   * Public method of type Boolean called "validateValue".
   * The method takes an Object parameter called "value".
   * The method loops through and checks to see if value is empty.
   * If so, it will return that the value is not null and that it is not empty.
   * Otherwise, it will check to see the value is an instance of Map.
   * If so, it will return that the value is not null and that it is not empty.
   * Otherwise, it will check to see if value is an instance of a String.
   * If so, it will return that the value is not null and that it is not empty.
   * Otherwise, it will check to see if value is an instance of a Collection Object.
   * If so, it will return that the value is not null and  if the value size is greater than zero.
   * Otherwise, it will just return that the value is not null and not empty.
   * @param <V>    - The type parameter
   * @param value  - Value to validate
   * @return       - Returns true if it is valid; false otherwise
   */
  @SuppressWarnings("rawtypes")
  public static <V> boolean validateValue(V value) {
    if (value == null) {
      return false;
    }
    if (value instanceof Collection) {
      return !((Collection)value).isEmpty();
    } else if (value instanceof Map){
      return !((Map)value).isEmpty();
    } else if (value instanceof Boolean) {
      return ((Boolean) value);
    }

    return !value.toString().trim().isEmpty();
  }

  /**
   * Validates all given values.
   * Public method of type Boolean called "validateValues".
   * The method takes multiple V type objects called "objectToValidate".
   * The method creates a boolean type variable called "valid" and sets it to true.
   * The method then loops through all the objects and validates each value in the object.
   * The method sets valid equal to the result of true and false.
   * If the method cannot validate the object values, it will throw an exception. If so, it returns false.
   * Otherwise, the method will return the variable called "valid".
   * @param <V>                - The type parameter
   * @param objectsToValidate  - Objects to validate
   * @return                   - Returns true if all objects are valid false otherwise.
   */
  @SafeVarargs
  public static <V> boolean validateValues(V... objectsToValidate) {
    return validateValues(false, toList(objectsToValidate));
  }

  @SafeVarargs
  public static <V> boolean validateValuesOr(V... objectsToValidate) {
    return validateValues(true, toList(objectsToValidate));
  }

  /**
   * Validates all given values to
   * @param or Set the or flag to check for at least one validated value
   * @param objectsToValidate objects to validate.
   * @param <V> the expected class type of the objects passed in. All objects need to be of same type
   * @return true if the values are valid, false otherwise
   */
  public static <V> boolean validateValues(boolean or, List<V> objectsToValidate) {
    boolean valid = !or;
    for (Object obj : objectsToValidate) {
      if (or && validateValue(obj)) {
        return true;
      }
      valid &= validateValue(obj);
    }

    return valid;
  }
  
  public static boolean isJUnitRunning() {
    StackTraceElement[] stackTrace = Thread.currentThread().getStackTrace();
    Optional<String> junit = StreamUtils.asStream(stackTrace)
        .map(StackTraceElement::getClassName)
        .filter(className -> className.startsWith("org.junit."))
        .findFirst();
    return junit.isPresent();
  }

  private static Set<Class<?>> getWrapperTypes() {
    return Set.of(
      Boolean.class,
      Character.class,
      Byte.class,
      Short.class,
      Integer.class,
      Long.class,
      Float.class,
      Double.class,
      String.class
    );
  }
}
