package net.locusworks.test;

import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Properties;

import net.locusworks.common.configuration.PropertiesManager;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.Test;

import static org.junit.jupiter.api.Assertions.*;

/**
 * Test cases for the properties manager class
 * @author Isaac Parenteau
 * @since 1.0.0-RELEASE
 *
 */
public class PropertiesManagerTest {

  private static final String PROPERTIES_FILE = "test.properties";
  private static final String TMP_PROPS = "temp.properties";
  private static final int ENTRY_SIZE = 4;
  
  public static enum Configuration {
    DB_HOST("dbHost"),
    DB_PORT("dbPort"),
    USER_EXPIRATION_DAYS("userExpirationDays"),
    LOG_LEVEL("logLevel");

    private final String value;

    private Configuration(String value) {
      this.value = value; 
    }

    /**
     * Get the current value of the enumeration
     * @return value
     */
    public String getValue() {
      return this.value;
    }

    @Override
    public String toString() {
      return this.value;
    }
  }

  @AfterAll
  public static void removeSavedProps() {
    File tmp = new File(TMP_PROPS);
    if (tmp.exists()) {
      tmp.delete();
    }
  }

  @Test
  public void testPropertiesLoad() {
    try {
      Properties props = PropertiesManager.loadConfiguration(this.getClass(), PROPERTIES_FILE);
      assertNotNull(props);
      assertTrue(props.containsKey(Configuration.USER_EXPIRATION_DAYS.toString()));
      assertTrue(props.containsKey(Configuration.DB_HOST.toString()));
      assertTrue(props.containsKey(Configuration.DB_PORT.toString()));
      assertTrue(props.containsKey(Configuration.LOG_LEVEL.toString()));
    } catch (IOException e) {
      fail(e.getMessage());
    }
  }

  @Test
  public void testAddConfiguration() {
    try {
      Properties props = PropertiesManager.loadConfiguration(this.getClass(), PROPERTIES_FILE);
      Properties tmp = new Properties();
      assertEquals(0, tmp.keySet().size());
      PropertiesManager.addConfiguration(tmp, props);
      assertEquals(ENTRY_SIZE, tmp.keySet().size());
      assertTrue(tmp.containsKey(Configuration.USER_EXPIRATION_DAYS.toString()));
      assertTrue(tmp.containsKey(Configuration.DB_HOST.toString()));
      assertTrue(tmp.containsKey(Configuration.DB_PORT.toString()));
      assertTrue(tmp.containsKey(Configuration.LOG_LEVEL.toString()));
    } catch (IOException e) {
      fail(e.getMessage());
    }
  }

  @Test 
  public void testRemoveConfiguration() {
    try {
      Properties props = PropertiesManager.loadConfiguration(this.getClass(), PROPERTIES_FILE);
      Properties tmp = new Properties();
      assert props != null;
      assertEquals(ENTRY_SIZE, props.keySet().size());
      assertEquals(0, tmp.keySet().size());
      PropertiesManager.removeConfiguration(props, tmp);
      assertEquals(0, props.keySet().size());
      assertEquals(0, tmp.keySet().size());
    } catch (IOException e) {
      fail(e.getMessage());
    }
  }

  @Test 
  public void testSaveConfiguration() {
    try {
      Properties props = PropertiesManager.loadConfiguration(this.getClass(), PROPERTIES_FILE);
      Path tmpFile = Paths.get(TMP_PROPS);
      PropertiesManager.saveConfiguration(props, tmpFile, "test propertis");
      Properties tmp = PropertiesManager.loadConfiguration(tmpFile);
      assertEquals(ENTRY_SIZE, tmp.keySet().size());
      assertTrue(tmp.containsKey(Configuration.USER_EXPIRATION_DAYS.toString()));
      assertTrue(tmp.containsKey(Configuration.DB_HOST.toString()));
      assertTrue(tmp.containsKey(Configuration.DB_PORT.toString()));
      assertTrue(tmp.containsKey(Configuration.LOG_LEVEL.toString()));
    } catch (IOException e) {
      fail(e.getMessage());
    }
  }
}
